from .args import arg, bool_or, json_value
from .command import command, subcommand
from .util import abort, confirm, printer

__version__ = "1.0a73"

__all__ = [
    "__version__",
    "abort",
    "arg",
    "bool_or",
    "command",
    "confirm",
    "json_value",
    "printer",
    "subcommand",
]



import shlex
import sys
from pathlib import Path

import rich.markup

from ..args import arg
from ..command import command
from ..result import Result
from ..util import flatten_args, isatty, printer, prompt, StreamOptions
from .local import local


@command
def remote(
    cmd: arg(container=list),
    host,
    user=None,
    port=None,
    sudo=False,
    run_as=None,
    sudo_prompt=False,
    shell="/bin/sh",
    cd=None,
    environ: arg(container=dict) = None,
    paths=(),
    # Args passed through to local command:
    stdout: arg(type=StreamOptions) = None,
    stderr: arg(type=StreamOptions) = None,
    echo=False,
    raise_on_error=True,
    dry_run=False,
    use_shared_connection=False,
) -> Result:
    """Run a remote command via SSH.

    Runs a remote shell command using ``ssh`` in a subprocess like so::

        ssh -q [-t] [<user>@]<host> [sudo [-u <run_as>] -H] /bin/sh -c '
            [cd <cd> &&]
            [export XYZ="xyz" &&]
            [export PATH="<path>" &&]
            <cmd>
        '

    Args:
        cmd (list|str): The command to run. If this is a list, it will
            be flattened into a string.
        host (str): Remote host to SSH into.
        user (str): Remote user to log in as (defaults to current local
            user).
        port (int): SSH port on remote host.
        sudo (bool): Run the remote command as root using ``sudo``.
        run_as (str): Run the remote command as a different user using
            ``sudo -u <run_as>``.
        sudo_prompt (bool): When set, a prompt for the sudo password
            will be shown on the *local* machine. The main use case for
            this is when the ``user`` requires a password to use sudo on
            the remote host. For passwordless sudo, this isn't needed.
        shell (str): The remote user's default shell will be used to run
            the remote command unless this is set to a different shell.
        cd (str): Where to run the command on the remote host.
        environ (dict): Extra environment variables to set on the remote
            host.
        paths (list): Additional paths to prepend to the remote
            ``$PATH``.
        stdout: See :obj:`runcommands.commands.local`.
        stderr: See :obj:`runcommands.commands.local`.
        echo: See :obj:`runcommands.commands.local`.
        raise_on_error: See :obj:`runcommands.commands.local`.
        dry_run: See :obj:`runcommands.commands.local`.
        use_shared_connection: Use a shared SSH connection instead of
            creating a new connection for every command run over SSH.
            NOTE: This should be considered experimental for now.

    """
    # Process Options --------------------------------------------------

    if not isinstance(cmd, str):
        cmd = flatten_args(cmd, join=True)

    tty_arg = "-t" if isatty(sys.stdin) else None
    port_args = ("-p", port) if port else ()
    ssh_connection_str = f"{user}@{host}" if user else host
    using_sudo = sudo or run_as

    if stdout:
        stdout = StreamOptions[stdout] if isinstance(stdout, str) else stdout

    # Build Remote Command ---------------------------------------------

    remote_cmd: list[str] = []

    if sudo:
        remote_cmd.extend(("sudo", "-H"))
    elif run_as:
        remote_cmd.extend(("sudo", "-H", "-u", run_as))

    if using_sudo and sudo_prompt:
        remote_cmd.extend(("--prompt=", "--stdin"))

    remote_cmd.extend((shell, "-c"))

    inner_cmd = []

    if cd:
        inner_cmd.append(f"cd {cd}")

    if environ:
        inner_cmd.extend(f'export {k}="{v}"' for k, v in environ.items())

    if paths:
        paths_str = ":".join(paths)
        inner_cmd.append(f'export PATH="{paths_str}:$PATH"')

    inner_cmd.append(cmd)
    inner_cmd_str = " &&\n    ".join(inner_cmd)
    inner_cmd_str = f"\n    {inner_cmd_str}\n"
    inner_cmd_str = shlex.quote(inner_cmd_str)

    remote_cmd.append(inner_cmd_str)
    remote_cmd_str = " ".join(remote_cmd)

    if using_sudo and sudo_prompt:
        sudo_prompt_message = rich.markup.escape("[sudo] password")
        sudo_password = prompt(sudo_prompt_message, password=True)
    else:
        sudo_password = None

    # Start Shared SSH Connection --------------------------------------
    #
    # TODO: Extract to utility function

    if use_shared_connection:
        try:
            home = Path.home()
        except RuntimeError:
            use_shared_connection = False
            printer.warning(
                "Could not create shared SSH connection: "
                "could not determine HOME directory"
            )
        else:
            socket_path = home / ".ssh" / f"runcommands-control-{user}-{host}-{port}"
            port_args = ("-p", port) if port else ()
            shared_connection_args = (
                "-S",
                socket_path,
                "-o",
                "ControlPersist=2m",
                tty_arg,
                port_args,
                ssh_connection_str,
            )
            check_result = local(
                ("ssh", "-O", "check", shared_connection_args),
                stderr=StreamOptions.hide,
                raise_on_error=False,
            )
            if check_result.failed:
                printer.info(f"Starting shared SSH connection: {socket_path}")
                local(("ssh", "-MN", shared_connection_args))
            else:
                printer.info(f"Reusing shared SSH connection: {socket_path}")

    # Run Remote Command -----------------------------------------------

    if use_shared_connection:
        tty_arg = None

    args = ("ssh", tty_arg, port_args, ssh_connection_str, remote_cmd_str)

    return local(
        args,
        input=sudo_password,
        stdout=stdout,
        stderr=stderr,
        echo=echo,
        raise_on_error=raise_on_error,
        dry_run=dry_run,
    )



from typing import MutableMapping

from .command import Command


class Collection(MutableMapping):
    """A collection of commands."""

    def __init__(self, commands):
        self.commands = commands

    @classmethod
    def load_from_module(cls, module):
        commands = {
            obj.name: obj
            for name, obj in vars(module).items()
            if isinstance(obj, Command) and not name.startswith("_")
        }
        return cls(commands)

    def set_attrs(self, **attrs):
        """Set the given attributes on *all* commands in collection."""
        commands = tuple(self.values())
        for name, value in attrs.items():
            for command in commands:
                setattr(command, name, value)

    def set_default_args(self, default_args):
        """Set default args for commands in collection.

        Default args are used when the corresponding args aren't passed
        on the command line or in a direct call.

        """
        for name, args in default_args.items():
            command = self[name]
            command.default_args = default_args.get(command.name) or {}

    def __getitem__(self, name):
        commands = self.commands
        if name in commands:
            return commands[name]
        name = Command.normalize_name(name)
        return commands[name]

    def __setitem__(self, name, command):
        self.commands[name] = command

    def __delitem__(self, name):
        del self.commands[name]

    def __iter__(self):
        return iter(self.commands)

    def __len__(self):
        return len(self.commands)



from rich.prompt import Confirm as BaseConfirm, Prompt, InvalidResponse

from .misc import abort
from .printer import printer


def prompt(message, password=False, choices=None, default=None, color=True):
    if color is True:
        color = "warning"
    if color:
        message = printer.colorize(message, color=color)
    try:
        return Prompt.ask(message, password=password, choices=choices, default=default)
    except KeyboardInterrupt:
        print()
        abort()


def confirm(
    prompt="Really?",
    color="warning",
    yes_value="y",
    abort_on_unconfirmed=False,
    abort_options=None,
):
    """Prompt for confirmation.

    Confirmation can be aborted by typing in a no value instead of one
    of the yes values or with Ctrl-C.

    Args:
        prompt (str): Prompt to present user ["Really?"]
        color (string|Color|bool) Color to print prompt string; can be
            ``False`` or ``None`` to print without color ["yellow"]
        yes_value (str): Value user must type in to confirm; note that
            this will be case sensitive *if* it contains any upper case
            letters ["y"]
        abort_on_unconfirmed (bool|int|str): When user does *not*
            confirm:

            - If this is an integer, print "Aborted" to stdout if
              it's 0 or to stderr if it's not 0 and then exit with
              this code
            - If this is a string, print it to stdout and exit with
              code 0
            - If this is ``True`` (or any other truthy value), print
              "Aborted" to stdout and exit with code 0

        abort_options (dict): Options to pass to :func:`abort` when not
            confirmed (these options will override any options set via
            ``abort_on_unconfirmed``)

    """
    if color:
        prompt = printer.colorize(prompt, color=color)

    choices = [yes_value, "n"]

    try:
        confirmed = Confirm.ask(prompt, choices=choices, default=False)
    except KeyboardInterrupt:
        print()
        confirmed = False

    # NOTE: The abort-on-unconfirmed logic is somewhat convoluted
    #       because of the special case for return code 0.

    do_abort_on_unconfirmed = not confirmed and (
        # True, non-zero return code, non-empty string, or any other
        # truthy value (in the manner of typical Python duck-typing)
        bool(abort_on_unconfirmed)
        or
        # Zero return code (special case)
        (abort_on_unconfirmed == 0 and abort_on_unconfirmed is not False)
    )

    if do_abort_on_unconfirmed:
        if abort_options is None:
            abort_options = {}

        if abort_on_unconfirmed is True:
            abort_options.setdefault("return_code", 0)
        elif isinstance(abort_on_unconfirmed, int):
            abort_options.setdefault("return_code", abort_on_unconfirmed)
        elif isinstance(abort_on_unconfirmed, str):
            abort_options.setdefault("message", abort_on_unconfirmed)
        else:
            abort_options.setdefault("return_code", 0)

        abort(**abort_options)

    return confirmed


class Confirm(BaseConfirm):
    @property
    def validate_error_message(self):
        yes, no = self.choices
        return f"[prompt.invalid]Please enter {yes} or {no}"

    def render_default(self, default):
        """Default is *always* no."""
        return "(n)"

    def process_response(self, value):
        value = value.strip()
        # Hitting enter without a value -> "n" -> unconfirmed
        if not value:
            return False
        yes, no = self.choices
        # If yes value contains *any* upper case characters, do case-
        # sensitive yes value comparison
        all_lower = all(c.islower() for c in yes)
        if all_lower:
            value = value.lower()
        if value == yes:
            return True
        value = value.lower()
        # Allow "yes" when yes value is lower case "y"
        if yes == "y" and value == "yes":
            return True
        # Always allow "no"
        if value in ("n", "no"):
            return False
        raise InvalidResponse(self.validate_error_message)



import enum
import subprocess


class Color(enum.Enum):
    default = "default"
    black = "black"
    red = "bright_red"
    green = "bright_green"
    yellow = "bright_yellow"
    blue = "bright_blue"
    magenta = "bright_magenta"
    cyan = "bright_cyan"
    white = "white"


class PrinterColor(enum.Enum):
    # Color
    default = Color.default.value
    black = Color.black.value
    red = Color.red.value
    green = Color.green.value
    yellow = Color.yellow.value
    blue = Color.blue.value
    magenta = Color.magenta.value
    cyan = Color.cyan.value
    white = Color.white.value

    # Colors by type
    none = Color.default.value
    header = Color.white.value
    info = Color.blue.value
    success = Color.green.value
    echo = Color.cyan.value
    warning = Color.yellow.value
    error = Color.red.value
    danger = Color.red.value
    debug = Color.cyan.value

    def open(self):
        return f"[{self.value}]"

    def close(self):
        return f"[/{self.value}]"


class StreamOptions(enum.Enum):
    """Choices for stream handling."""

    capture = "capture"
    hide = "hide"
    none = "none"

    @property
    def option(self):
        return {
            "capture": subprocess.PIPE,
            "hide": subprocess.DEVNULL,
            "none": None,
        }[self.value]



import os
import sys
from typing import Any

import toml
from importlib import import_module
from pathlib import Path

from . import __version__
from .args import arg, json_value
from .command import Command
from .collection import Collection
from .exc import RunAborted, RunnerError
from .runner import CommandRunner
from .util import (
    abs_path,
    is_mapping,
    is_project_root,
    is_sequence,
    merge_dicts,
    module_from_path,
    printer,
)


class Run(Command):
    name = "runcommands"

    allowed_config_file_args = (
        "globals",
        "envs",
        "args",
        "environ",
    )

    def implementation(
        self,
        commands_module: arg(
            short_option="-m",
        ) = None,
        config_file: arg(
            short_option="-f",
        ) = None,
        # Globals
        globals_: arg(
            container=dict,
            type=json_value,
            help="Global variables & default args for *all* commands; will be "
            "injected into itself, default args, and environment variables "
            "(higher precedence than keyword args)",
        ) = None,
        # Special globals (for command line convenience)
        env: arg(
            help="env will be added to globals if specified",
        ) = None,
        version: arg(
            help="version will be added to globals if specified",
        ) = None,
        echo: arg(
            type=bool,
            help="echo=True will be added to globals",
            inverse_help="echo=False will be added to globals",
        ) = None,
        # Environment variables
        environ: arg(
            container=dict,
            help="Additional environment variables; added just before commands are run",
        ) = None,
        # Meta
        info: arg(
            no_inverse=True,
            help="Show info and exit",
            mutual_exclusion_group="meta-show",
        ) = False,
        list_commands: arg(
            no_inverse=True,
            help="Show info & commands and exit",
            mutual_exclusion_group="meta-show",
        ) = False,
        debug: arg(
            type=bool,
            help="Print debugging info & re-raise exceptions; also added to globals",
        ) = None,
        *,
        all_argv=(),
        run_argv=(),
        command_argv=(),
    ):
        """Run one or more commands in succession.

        For example, assume the commands ``local`` and ``remote`` have been
        defined; the following will run ``ls`` first on the local host and
        then on the remote host::

            runcommands local ls remote <host> ls

        When a command name is encountered in ``argv``, it will be considered
        the starting point of the next command *unless* the previous item in
        ``argv`` was an option like ``--xyz`` that expects a value (i.e.,
        it's not a flag).

        To avoid ambiguity when an option value matches a command name, the
        value can be prepended with a colon to force it to be considered
        a value and not a command name.

        """
        commands_module = self.find_commands_module(commands_module)

        if commands_module is None:
            raise RunnerError("Could not find commands module")

        config_file = self.find_config_file(config_file)
        collection = Collection.load_from_module(commands_module)
        cli_globals = globals_ or {}

        if env:
            cli_globals["env"] = env
        if version:
            cli_globals["version"] = version
        if echo is not None:
            cli_globals["echo"] = echo
        if debug is not None:
            cli_globals["debug"] = debug

        if config_file:
            args_from_file = self.read_config_file(config_file, collection)
            args = merge_dicts(args_from_file, {"environ": environ or {}})
            config_file_globals = args["globals"]
            env = env or config_file_globals.get("env")

            if env:
                envs = args["envs"]

                try:
                    env_globals = envs[env]
                except KeyError:
                    raise RunnerError(f"Unknown env: {env}")

                # Don't add the selected env's default args dict to
                # globals, but save it so it can be added back to the
                # the global envs dict (for inspection purposes).
                env_default_args = env_globals.pop("args")

                globals_ = merge_dicts(config_file_globals, env_globals, cli_globals)
                globals_["envs"] = envs

                env_globals["args"] = env_default_args
            else:
                env_default_args = {}
                globals_ = merge_dicts(config_file_globals, cli_globals)

            base_default_args = merge_dicts(args["args"], env_default_args)

            default_args: dict[str, dict[str, Any]] = {name: {} for name in collection}
            default_args = merge_dicts(default_args, base_default_args)

            for command_name, command_default_args in default_args.items():
                command = collection[command_name]

                # Add globals that correspond to this command (that
                # aren't present in default args section).
                for name, value in globals_.items():
                    param = command.find_parameter(name)
                    if param is not None:
                        if param.name not in command_default_args:
                            command_default_args[param.name] = value
                    elif command.has_kwargs:
                        name = Command.normalize_name(name)
                        command_default_args[name] = value

                if "default_args" not in default_args:
                    # This gives commands access to both their own and
                    # other commands' default args.
                    if command.find_parameter("default_args"):
                        command_default_args["default_args"] = base_default_args

                # Convert lists to tuples for the command's args that are
                # specified as being tuples.
                for name, value in command_default_args.items():
                    command_arg = command.find_arg(name)
                    if (
                        command_arg
                        and command_arg.container
                        and isinstance(value, list)
                    ):
                        command_default_args[name] = command_arg.container(value)

            default_args = {name: args for name, args in default_args.items() if args}

            environ = args["environ"]
        else:
            globals_ = cli_globals
            default_args = {}
            environ = environ or {}

        debug = globals_.get("debug", False)
        show_info = info or list_commands or not command_argv or debug
        print_and_exit = info or list_commands

        globals_, default_args, environ = self.interpolate(
            globals_, default_args, environ
        )

        if show_info:
            print("RunCommands", __version__)

        if debug:
            printer.debug("Commands module:", commands_module)
            printer.debug("Config file:", config_file)
            printer.debug("All args:", all_argv)
            printer.debug("Run args:", run_argv)
            printer.debug("Command args:", command_argv)
            items = (
                ("Globals:", globals_),
                ("Default args:", default_args),
                ("Env default args:", env_default_args),
                ("Environment variables:", environ),
            )
            for label, data in items:
                if data:
                    printer.debug(label)
                    for k in sorted(data):
                        v = data[k]
                        printer.debug(f"  - {k} = {v!r}")
                else:
                    printer.debug(label, data)

        if environ:
            os.environ.update(environ)

        collection.set_attrs(debug=debug)
        collection.set_default_args(default_args)
        runner = CommandRunner(collection, debug)

        if print_and_exit:
            if list_commands:
                runner.print_usage()
        elif not command_argv:
            printer.warning("\nNo command(s) specified")
            runner.print_usage()
        else:
            runner.run(command_argv)

    def run(self, argv, **kwargs):
        all_argv, run_argv, command_argv = self.partition_argv(argv)
        if "-d" in run_argv or "--debug" in run_argv:
            self.debug = True
        kwargs.update(
            {
                "all_argv": all_argv,
                "run_argv": run_argv,
                "command_argv": command_argv,
            }
        )
        return super().run(run_argv, **kwargs)

    def console_script(self, argv=None, **overrides):
        _argv = sys.argv[1:] if argv is None else argv
        if "-d" in _argv or "--debug" in _argv:
            self.debug = True
        return super().console_script(argv, **overrides)

    def partition_argv(self, argv=None):
        if argv is None:
            argv = sys.argv[1:]

        if not argv:
            return argv, [], []

        # Consume all args that appear to be options (and their values,
        # if applicable), even those that aren't know run options, up
        # until the first non-option word is reached. That word is
        # assumed to be the start of commands.

        def looks_like_option(s):
            return bool(
                (s.startswith("-") or s.startswith("--"))
                and not s.startswith("---")
                and s.strip("-")
            )

        i = 0
        argc = len(argv)
        run_argv = []
        parse_optional = self.parse_optional
        parse_multi_short_option = self.parse_multi_short_option

        while i < argc:
            a = argv[i]

            if a == "--":
                # Explicit end of run args.
                i += 1
                break

            option_data = parse_optional(a)

            if option_data is not None:
                # Arg is a known run option.
                name, option, value = option_data
                run_argv.append(a)

                if a in ("-d", "--debug"):
                    self.debug = True
                elif value is None and option.takes_value:
                    # Collect the option's value if it takes one and one
                    # wasn't provided via --opt=<value>.
                    j = i + 1
                    if j < argc:
                        run_argv.append(argv[j])
                        i = j
            else:
                if not looks_like_option(a):
                    # Non-option word; assumed to be start of commands.
                    break

                short_options, value = parse_multi_short_option(a)

                if short_options is None:
                    run_argv.append(a)
                else:
                    run_argv.extend(short_options)

                    if "-d" in short_options:
                        self.debug = True

                    if value is not None:
                        run_argv.append(value)
                    else:
                        # Collect the last short option's value if it
                        # takes one and one wasn't provided via
                        # -abc<value>.
                        option_data = parse_optional(short_options[-1])
                        if option_data is not None:
                            name, option, value = option_data
                            if value is None and option.takes_value:
                                j = i + 1
                                if j < argc:
                                    run_argv.append(argv[j])
                                    i = j

            i += 1

        command_argv = argv[i:]
        return argv, run_argv, command_argv

    def find_commands_module(self, commands_module, start_dir="."):
        if commands_module:
            if commands_module.endswith(".py"):
                commands_module = abs_path(commands_module)
                if not os.path.isfile(commands_module):
                    raise RunnerError(
                        f"Commands file does not exist: {commands_module}"
                    )
                else:
                    return module_from_path("commands", commands_module)
            else:
                try:
                    return import_module(commands_module)
                except ImportError:
                    raise RunnerError(
                        f"Commands module could not be imported: {commands_module}"
                    )
        current_dir = Path(start_dir).resolve()
        file_system_root = Path(current_dir.root)
        checked_file_system_root = False
        candidates = ("runcommands.py", "commands.py")
        while not checked_file_system_root:
            for candidate in candidates:
                candidate = current_dir / candidate
                if candidate.is_file():
                    return module_from_path("commands", candidate)
            if current_dir == file_system_root:
                checked_file_system_root = True
            if is_project_root(current_dir):
                break
            current_dir = current_dir.parent
        return None

    def find_config_file(self, config_file, start_dir="."):
        if config_file:
            config_file = abs_path(config_file)
            if not os.path.exists(config_file):
                raise RunnerError(f"Config file does not exists: {config_file}")
            return config_file
        current_dir = Path(start_dir).resolve()
        file_system_root = Path(current_dir.root)
        checked_file_system_root = False
        candidates = ("runcommands.toml", "commands.toml", "pyproject.toml")
        while not checked_file_system_root:
            for candidate in candidates:
                candidate = current_dir / candidate
                if candidate.is_file():
                    return candidate
            if current_dir == file_system_root:
                checked_file_system_root = True
            if is_project_root(current_dir):
                break
            current_dir = current_dir.parent
        return None

    def read_config_file(self, config_file, collection):
        return self._read_config_file(config_file, collection)

    def _read_config_file(self, config_file, collection):
        with open(config_file) as fp:
            args = toml.load(fp)

        if os.path.basename(config_file) == "pyproject.toml":
            tool = args.get("tool") or {}
            args = tool.get("runcommands") or {}

        for name in self.allowed_config_file_args:
            # Not present or present but not set
            if args.get(name) is None:
                args[name] = {}

        extends = args.pop("extends", None)

        for name in tuple(args):
            if name not in self.allowed_config_file_args:
                raise RunnerError(f"Arg cannot be specified in config file: {name}")

        self.normalize_command_and_arg_names(args["args"], collection)

        envs = args["envs"]
        for env, data in envs.items():
            if data is None:
                data = {}
                envs[env] = data
            data.setdefault("args", {})
            self.normalize_command_and_arg_names(data["args"], collection)

        if extends:
            extends = abs_path(extends, relative_to=os.path.dirname(config_file))
            extended_args = self._read_config_file(extends, collection)
            args = merge_dicts(extended_args, args)

        return args

    def normalize_command_and_arg_names(self, config, collection):
        for command_name in tuple(config):
            try:
                command = collection[command_name]
            except KeyError:
                raise RunnerError(
                    f"Unknown command in default args section of "
                    f"config file: {command_name}"
                )

            if command.name != command_name:
                config[command.name] = config.pop(command_name)

            command_default_args = config[command.name]
            for name in tuple(command_default_args):
                param = command.find_parameter(name)
                if param is None:
                    raise RunnerError(
                        f"Unknown arg for command {command_name} in "
                        f"default args section of config file: {name}"
                    )
                if param is not None and name != param.name:
                    command_default_args[param.name] = command_default_args.pop(name)

    def interpolate(self, globals_, default_args, environ):
        if globals_:
            globals_ = self._interpolate(globals_, globals_)

        if default_args:
            context = merge_dicts(globals_, default_args)
            default_args = self._interpolate(default_args, context)

        if environ:
            environ = self._interpolate(environ, globals_)

        return globals_, default_args, environ

    def _interpolate(self, obj, context):
        if is_mapping(obj):
            items = ((k, self._interpolate(v, context)) for k, v in obj.items())
            obj = obj.__class__(items)
        elif is_sequence(obj):
            items = (self._interpolate(v, context) for v in obj)
            obj = obj.__class__(items)
        else:
            obj = self._inject(obj, context)
        return obj

    def _inject(self, value, context, start=0):
        if not isinstance(value, str):
            return value
        i = value.find("{{", start)
        if i == -1:
            return value
        h = i - 1
        if h >= 0 and value[h] == "\\":
            return self._inject(value, context, h + 2)
        j = value.rfind("}}", i + 2)
        if j == -1:
            # String looks like "{{ abc"
            # XXX: Error?
            if self.debug:
                printer.warning(f'Unclosed interpolation group in value: "{value}"')
            return value
        k = j + 2
        key = value[i + 2 : j].strip()
        if not key:
            # String looks like "{{}} xyz"
            # XXX: Error?
            if self.debug:
                printer.warning(f'Empty interpolation group in value: "{value}"')
            return value
        context_value = self._find_in_context(context, key)
        context_value = self._inject(context_value, context)
        if i == 0 and k == len(value):
            value = context_value
        else:
            value = f"{value[:i]}{context_value}{value[k:]}"
            value = self._inject(value, context)
        return value

    def _find_in_context(self, context, key):
        value = context
        parts = key.split(".")
        for segment in parts:
            value = value[segment]
        return value

    def sigint_handler(self, _sig_num, _frame):
        raise RunAborted(0, message="\nAborted by Ctrl-C (SIGINT)")


run = Run()



import re


def camel_to_underscore(name):
    """Convert camel case name to underscore name.

    Examples::

        >>> camel_to_underscore('HttpRequest')
        'http_request'
        >>> camel_to_underscore('httpRequest')
        'http_request'
        >>> camel_to_underscore('HTTPRequest')
        'http_request'
        >>> camel_to_underscore('myHTTPRequest')
        'my_http_request'
        >>> camel_to_underscore('MyHTTPRequest')
        'my_http_request'
        >>> camel_to_underscore('my_http_request')
        'my_http_request'
        >>> camel_to_underscore('MyHTTPRequestXYZ')
        'my_http_request_xyz'
        >>> camel_to_underscore('_HTTPRequest')
        '_http_request'
        >>> camel_to_underscore('Request')
        'request'
        >>> camel_to_underscore('REQUEST')
        'request'
        >>> camel_to_underscore('_Request')
        '_request'
        >>> camel_to_underscore('__Request')
        '__request'
        >>> camel_to_underscore('_request')
        '_request'
        >>> camel_to_underscore('Request_')
        'request_'

    """
    name = re.sub(r"(?<!\b)(?<!_)([A-Z][a-z])", r"_\1", name)
    name = re.sub(r"(?<!\b)(?<!_)([a-z])([A-Z])", r"\1_\2", name)
    name = name.lower()
    return name


def invert_string(string):
    """Invert the logical meaning of a string.

    Examples::

        >>> invert_string('Yes')
        'No'
        >>> invert_string('No')
        'Yes'
        >>> invert_string("don't")
        'do'
        >>> invert_string("don't do thing")
        'do thing'

    """
    first_letter = string[0]
    is_capitalized = first_letter.isupper()
    words = string.split(None, 1)
    first_word = words[0]

    one_word_inversions = {
        "Do": "Don't",
        "Don't": "Do",
        "Do not": "Do",
        "With": "Without",
        "Without": "With",
        "Yes": "No",
        "No": "Yes",
    }
    for k in tuple(one_word_inversions):
        one_word_inversions[k.lower()] = one_word_inversions[k].lower()

    inversions = {
        "Do": "No",
        "Don't": "",
        "Do not": "",
        "With": "Without",
        "Without": "With",
        "No": "With",
    }
    for k in tuple(inversions):
        inversions[k.lower()] = inversions[k].lower()

    if string in one_word_inversions:
        inverted = one_word_inversions[string]
    elif first_word in inversions:
        inverted_first_word = inversions[first_word]
        if inverted_first_word:
            words[0] = inverted_first_word
            i = len(first_word) - len(string)
            inverted = f"{inverted_first_word}{string[i:]}"
        else:
            inverse_help = words[1]
            inverted = inverse_help.capitalize() if is_capitalized else inverse_help
    else:
        inverted_first_word = "Don't" if is_capitalized else "don't"
        inverted = f"{inverted_first_word} {first_letter.lower()}{string[1:]}"

    suffixes = {
        "[y]": "[n]",
        "(y)": "(n)",
        "[yes]": "[no]",
        "(yes)": "(no)",
        "[true]": "[false]",
        "(true)": "(false)",
    }
    for suffix, inverse_suffix in suffixes.items():
        if inverted.endswith(suffix):
            inverted = f"{inverted[: -len(suffix)]}{inverse_suffix}"
        elif inverted.endswith(inverse_suffix):
            inverted = f"{inverted[: -len(inverse_suffix)]}{suffix}"

    return inverted



from functools import partial

from rich.console import Console

from .enums import PrinterColor


class Printer:
    colors: type[PrinterColor]
    stdout_console: Console
    stderr_console: Console

    def __init__(self, colors: type[PrinterColor] = PrinterColor):
        self.colors = colors
        self.stdout_console = Console()
        self.stderr_console = Console(stderr=True)

    def __call__(self, *args, **kwargs):
        self.print(*args, **kwargs)

    def __getattr__(self, color):
        # self.red("...")
        return partial(self.print, color=color)

    def get_color(self, color: PrinterColor | str | None) -> PrinterColor | None:
        if color is None:
            return None
        if isinstance(color, str):
            try:
                return self.colors[color]
            except KeyError:
                raise ValueError(f"Unknown color: {color}") from None
        return color

    def colorize(self, *args, color: PrinterColor | str | None = None, sep=" "):
        if not args:
            return ""

        color_entry = self.get_color(color)
        string = []

        if color_entry is not None:
            string.append(color_entry.open())

        for arg in args[:-1]:
            if isinstance(arg, PrinterColor):
                string.append(arg.open())
            else:
                string.append(str(arg))
                string.append(sep)

        string.append(str(args[-1]))

        if color_entry is not None:
            string.append(color_entry.close())

        return "".join(string)

    def print(
        self,
        *args,
        color=None,
        sep=" ",
        highlight=False,
        stderr=False,
        **kwargs,
    ):
        string = self.colorize(*args, color=color, sep=sep)
        if "flush" in kwargs:
            del kwargs["flush"]
        console = self.stderr_console if stderr else self.stdout_console
        console.print(string, sep=sep, highlight=highlight, **kwargs)

    def header(self, *args, color=None, style="bold", **kwargs):
        if color is None:
            color = self.colors.header
        self.hr()
        self.print(*args, color=color, style=style, **kwargs)
        self.print()

    def info(self, *args, color=None, **kwargs):
        if color is None:
            color = self.colors.info
        self.print(*args, color=color, **kwargs)

    def success(self, *args, color=None, **kwargs):
        if color is None:
            color = self.colors.success
        self.print(*args, color=color, **kwargs)

    def echo(self, *args, color=None, **kwargs):
        if color is None:
            color = self.colors.echo
        self.print(*args, color=color, **kwargs)

    def warning(self, *args, color=None, stderr=True, **kwargs):
        if color is None:
            color = self.colors.warning
        self.print(*args, color=color, stderr=stderr, **kwargs)

    def error(self, *args, color=None, stderr=True, **kwargs):
        if color is None:
            color = self.colors.error
        self.print(*args, color=color, stderr=stderr, **kwargs)

    def danger(self, *args, color=None, stderr=True, **kwargs):
        if color is None:
            color = self.colors.danger
        self.print(*args, color=color, stderr=stderr, **kwargs)

    def debug(self, *args, color=None, stderr=True, **kwargs):
        if color is None:
            color = self.colors.debug
        self.print(*args, color=color, stderr=stderr, **kwargs)

    def hr(self, *args, color=None, fill_char="─", align="center", **kwargs):
        """Print horizontal rule with optional title"""
        kwargs["characters"] = fill_char
        kwargs["align"] = align
        if "end" in kwargs:
            end = kwargs.pop("end")
            args = args + (end,)
        if args:
            sep = kwargs.get("sep") or " "
            kwargs["title"] = sep.join(args)
        if color:
            kwargs["style"] = self.get_color(color).value
        self.stdout_console.rule(**kwargs)


printer = Printer()



import inspect
from collections import OrderedDict

from .data import Data
from .enums import Color, StreamOptions
from .misc import (
    abort,
    flatten_args,
    format_if,
    is_mapping,
    is_sequence,
    is_type,
    isatty,
    load_object,
    merge_dicts,
)
from .path import (
    abs_path,
    asset_path,
    find_project_root,
    is_project_root,
    module_from_path,
    paths_to_str,
)
from .printer import printer
from .prompt import confirm, prompt
from .string import camel_to_underscore, invert_string


__all__ = [
    "get_commands_in_namespace",
    "Data",
    "Color",
    "StreamOptions",
    "abort",
    "flatten_args",
    "format_if",
    "isatty",
    "is_mapping",
    "is_sequence",
    "is_type",
    "load_object",
    "merge_dicts",
    "abs_path",
    "asset_path",
    "find_project_root",
    "is_project_root",
    "module_from_path",
    "paths_to_str",
    "printer",
    "confirm",
    "prompt",
    "camel_to_underscore",
    "invert_string",
]


def get_commands_in_namespace(namespace=None, level=1):
    """Get commands in namespace.

    Args:
        namespace (dict|module): Typically a module. If not passed, the
            globals from the call site will be used.
        level (int): If not called from the global scope, set this
            appropriately to account for the call stack.

    Returns:
        OrderedDict: The commands found in the namespace, ordered by
            name.

    Can be used to create ``__all__`` lists::

        __all__ = list(get_commands_in_namespace())

    """
    from ..command import Command  # noqa: Avoid circular import

    commands = {}
    if namespace is None:
        frame = inspect.stack()[level][0]
        namespace = frame.f_globals
    elif inspect.ismodule(namespace):
        namespace = vars(namespace)
    for name in namespace:
        obj = namespace[name]
        if isinstance(obj, Command):
            commands[name] = obj
    return OrderedDict((name, commands[name]) for name in sorted(commands))



import os
import shlex
from subprocess import CompletedProcess
from typing import Any, Mapping

from cached_property import cached_property

from .exc import RunCommandsError


class Result(RunCommandsError):
    args: tuple[Any]
    return_code: int
    stdout: str
    stderr: str
    succeeded: bool
    failed: bool

    def __init__(self, args: str | list[Any] | tuple[Any], return_code, stdout, stderr):
        args = shlex.split(args) if isinstance(args, str) else args
        self.args = tuple(args)
        self.return_code = return_code
        self.stdout = stdout
        self.stderr = stderr
        self.succeeded = self.return_code == 0
        self.failed = not self.succeeded

    @classmethod
    def from_subprocess_result(cls, result: CompletedProcess[Any]):
        return cls(
            result.args,
            result.returncode,
            result.stdout,
            result.stderr,
        )

    @cached_property
    def args_str(self):
        args = self.args
        if isinstance(args, Mapping):
            return " ".join(f"{k} => {v}" for k, v in args.items())
        # XXX: Assume list, tuple, or some other kind of sequence
        return " ".join(str(a) for a in args)

    @cached_property
    def stdout_lines(self):
        return self.stdout.splitlines() if self.stdout else []

    @cached_property
    def stderr_lines(self):
        return self.stderr.splitlines() if self.stderr else []

    def __bool__(self):
        return self.succeeded

    def __str__(self):
        output = (self.stderr if self.return_code else self.stdout) or "[NO OUTPUT]"
        if output.endswith(os.linesep):
            output = output[:-1]
        status = "SUCCEEDED" if self.succeeded else "FAILED"
        string = f"{status} ({self.return_code}): {self.args_str} -> {output}"
        return string

    def __repr__(self):
        return repr(str(self))



from ..args import arg
from ..command import command
from ..result import Result
from ..util import abs_path, StreamOptions
from .local import local


@command
def sync(
    source,
    destination,
    host,
    user=None,
    sudo=False,
    run_as=None,
    options=("-rltvz", "--no-perms", "--no-group"),
    excludes=(),
    exclude_from=None,
    delete=False,
    dry_run=False,
    mode="u=rwX,g=rwX,o=",
    quiet=True,
    pull=False,
    # Args passed through to local command:
    stdout: arg(type=StreamOptions) = None,
    stderr: arg(type=StreamOptions) = None,
    echo=False,
    raise_on_error=True,
) -> Result:
    """Sync files using rsync.

    By default, a local ``source`` is pushed to a remote
    ``destination``. To pull from a remote ``source`` to a local
    ``destination`` instead, pass ``pull=True``.

    """
    source = abs_path(source, keep_slash=True)
    destination = abs_path(destination, keep_slash=True)
    connection_str = f"{user}@{host}" if user else host
    push = not pull

    if sudo:
        rsync_path = ("--rsync-path", "sudo rsync")
    elif run_as:
        rsync_path = ("--rsync-path", f"sudo -u {run_as} rsync")
    else:
        rsync_path = None

    if push:
        destination = f"{connection_str}:{destination}"
    else:
        source = f"{connection_str}:{source}"

    args = (
        "rsync",
        rsync_path,
        options,
        ("--chmod", mode) if mode else None,
        tuple(("--exclude", exclude) for exclude in excludes),
        ("--exclude-from", exclude_from) if exclude_from else None,
        "--delete" if delete else None,
        "--dry-run" if dry_run else None,
        "--quiet" if quiet else None,
        source,
        destination,
    )
    return local(
        args, stdout=stdout, stderr=stderr, echo=echo, raise_on_error=raise_on_error
    )



import argparse
import builtins
import json
import re
from enum import Enum
from functools import update_wrapper
from inspect import Parameter as BaseParameter

from cached_property import cached_property

from .exc import CommandError
from .util import invert_string, is_mapping, is_sequence, is_type

EMPTY = BaseParameter.empty
KEYWORD_ONLY = BaseParameter.KEYWORD_ONLY
POSITIONAL_ONLY = BaseParameter.POSITIONAL_ONLY
POSITIONAL_OR_KEYWORD = BaseParameter.POSITIONAL_OR_KEYWORD
VAR_KEYWORD = BaseParameter.VAR_KEYWORD
VAR_POSITIONAL = BaseParameter.VAR_POSITIONAL


class PositionalPlaceholder:
    """Used as a placeholder for positionals."""


class Parameter:
    """Wrapper for :class:`inspect.Parameter`.

    Adds convenience methods for our typical use cases.

    """

    empty = EMPTY
    KEYWORD_ONLY = KEYWORD_ONLY
    POSITIONAL_ONLY = POSITIONAL_ONLY
    POSITIONAL_OR_KEYWORD = POSITIONAL_OR_KEYWORD
    VAR_KEYWORD = VAR_KEYWORD
    VAR_POSITIONAL = VAR_POSITIONAL

    def __init__(self, parameter):
        self.parameter = parameter

    @cached_property
    def is_positional(self):
        kind = self.parameter.kind
        default = self.parameter.default
        return (kind is POSITIONAL_ONLY) or (
            kind is POSITIONAL_OR_KEYWORD and default is EMPTY
        )

    @cached_property
    def is_var_positional(self):
        return self.parameter.kind is VAR_POSITIONAL

    @cached_property
    def is_var_keyword(self):
        return self.parameter.kind is VAR_KEYWORD

    @cached_property
    def is_optional(self):
        kind = self.parameter.kind
        default = self.parameter.default
        return (
            (kind is POSITIONAL_OR_KEYWORD) or (kind is KEYWORD_ONLY)
        ) and default is not EMPTY

    @cached_property
    def is_required_keyword_only(self):
        kind = self.parameter.kind
        default = self.parameter.default
        return kind is KEYWORD_ONLY and default is EMPTY

    @cached_property
    def is_bool(self):
        return isinstance(self.parameter.default, bool)

    def __getattr__(self, name):
        """Proxy to wrapped :class:`inspect.Parameter`."""
        return getattr(self.parameter, name)


class ArgConfig:
    """Configuration for an arg.

    This can be used as a function parameter annotation to explicitly
    configure an arg, overriding default behavior.

    Args:

        container (type): Container type to collect args into. If this
            is passed, or if the ``default`` value is a container type,
            values for this arg will be collected into a container of
            the appropriate type.
        type (type): The arg's type. By default, a positional arg will
            be parsed as ``str`` and an optional/keyword arg will be
            parsed as the type of its default value (or as ``str`` if
            the default value is ``None``). If a ``container`` is
            specified, or if the ``default`` value for the arg is a
            container, the ``type`` will be applied to the container's
            values.
        choices (sequence): A sequence of allowed choices for the arg.
        help (str): Help string for the arg.
        inverse_help (str): Inverse help string for the arg (for the
            ``--no-xyz`` variation of boolean args).
        short_option (str): Short command line option.
        long_option (str): Long command line option.
        inverse_short_option (str): Inverse short option for boolean
            args.
        inverse_long_option (str): Inverse long option for boolean args.
        action (Action): ``argparse`` Action.
        nargs (int|str): Number of command line args to consume.
        mutual_exclusion_group (str): Name of mutual exclusion group to
            add this arg to.
        envvar(str): Environment variable containing a default value to
            use when the arg isn't specified on the command line.
        default: Default value for positional args.

    .. note:: For convenience, regular dicts can be used to annotate
        args instead; they will be converted to instances of this class
        automatically.

    """

    short_option_regex = re.compile(r"-\w")
    long_option_regex = re.compile(r"--\w+(-\w+)*")

    def __init__(
        self,
        *,
        container=None,
        type=None,
        choices=None,
        help=None,
        inverse_help=None,
        short_option=None,
        long_option=None,
        no_inverse=False,
        inverse_short_option=None,
        inverse_long_option=None,
        inverse_option=None,  # XXX: Temporary alias for inverse_long_option
        action=None,
        nargs=None,
        mutual_exclusion_group=None,
        envvar=None,
        default=EMPTY,
    ):
        if short_option is not None:
            if not self.short_option_regex.fullmatch(short_option):
                raise CommandError(
                    f'Expected short option with form -x, not "{short_option}"'
                )

        if long_option is not None:
            if not self.long_option_regex.fullmatch(long_option):
                raise CommandError(
                    f'Expected long option with form --option, not "{long_option}"'
                )

        self.container = container
        self.type = type
        self.choices = choices
        self.help = help
        self.inverse_help = inverse_help
        self.short_option = short_option
        self.long_option = long_option
        self.no_inverse = no_inverse
        self.inverse_short_option = inverse_short_option
        self.inverse_long_option = inverse_long_option or inverse_option
        self.action = action
        self.nargs = nargs
        self.mutual_exclusion_group = mutual_exclusion_group
        self.envvar = envvar
        self.default = default

    def __repr__(self):
        type_name = self.type.__name__ if self.type is not None else "None"
        options = (
            self.short_option,
            self.long_option,
            self.inverse_short_option,
            self.inverse_long_option,
        )
        options = (option for option in options if option)
        options = ", ".join(options)
        return f"arg<{type_name}>({options})"


arg = ArgConfig


class Arg:
    """Encapsulates an arg belonging to a command.

    Attributes:

        command (Command): Command this arg belongs to.
        parameter (Parameter): Function parameter this arg is derived
            from.
        name (str): Normalized arg name.
        container (type): Container type to collect args into. If this
            is passed, or if the ``default`` value is a container type,
            values for this arg will be collected into a container of
            the appropriate type.
        type (type): The arg's type. By default, a positional arg will
            be parsed as ``str`` and an optional/keyword arg will be
            parsed as the type of its default value (or as ``str`` if
            the default value is ``None``). If a ``container`` is
            specified, or if the ``default`` value for the arg is a
            container, the ``type`` will be applied to the container's
            values.
        positional (bool): An arg will automatically be considered
            positional if it doesn't have a default value, so this
            doesn't usually need to be passed explicitly. It can be
            used to force an arg that would normally be optional to
            be positional.
        default (object): Default value for the arg.
        choices (sequence): A sequence of allowed choices for the arg.
        help (str): Help string for the arg.
        inverse_help (str): Inverse help string for the arg (for the
            ``--no-xyz`` variation of boolean args).
        short_option (str): Short command line option.
        long_option (str): Long command line option.
        inverse_short_option (str): Inverse short option for boolean
            args.
        inverse_long_option (str): Inverse long option for boolean args.
        action (Action): ``argparse`` Action.
        nargs (int|str): Number of command line args to consume.
        mutual_exclusion_group (str): Name of mutual exclusion group to
            add this arg to.
        envvar(str): Environment variable containing a default value to
            use when the arg isn't specified on the command line.

    """

    def __init__(
        self,
        *,
        command,
        parameter,
        name,
        container,
        type,
        positional,
        default,
        choices,
        help,
        inverse_help,
        short_option,
        long_option,
        no_inverse,
        inverse_short_option,
        inverse_long_option,
        action,
        nargs,
        mutual_exclusion_group,
        envvar,
    ):
        is_keyword_only = parameter.kind is KEYWORD_ONLY and default is EMPTY
        is_var_positional = parameter.kind is VAR_POSITIONAL

        if default is EMPTY:
            is_positional = not (is_var_positional or is_keyword_only)
            is_optional = False
        else:
            is_positional = False
            is_optional = True

        if positional is not None:
            is_positional = positional

        metavar = name.upper().replace("-", "_")
        if container and len(name) > 1 and name.endswith("s"):
            metavar = metavar[:-1]

        if container is None:
            if is_mapping(default) or is_sequence(default):
                container = default.__class__
            elif is_var_positional:
                container = tuple

        if type is None:
            if is_type(choices, Enum):
                type = choices
            elif container is not None:
                if default and default is not EMPTY:
                    type = default[0].__class__
                else:
                    type = str
            elif default not in (None, EMPTY):
                type = default.__class__
            else:
                type = str

        if isinstance(type, builtins.type):
            is_bool = issubclass(type, bool)
            is_bool_or = issubclass(type, bool_or)
            is_enum_bool_or = is_bool_or and issubclass(type, Enum)
            is_enum = issubclass(type, Enum)
        else:
            is_bool = False
            is_bool_or = False
            is_enum_bool_or = False
            is_enum = False

        if is_bool:
            type = None
            metavar = None
        elif is_bool_or:
            type = type.type

        if not choices:
            if is_enum:
                choices = type
            elif is_enum_bool_or:
                choices = type.type

        if is_positional or is_var_positional:
            options = (short_option, long_option, inverse_long_option)
            options = tuple(option for option in options if option is not None)
            if options:
                raise CommandError(
                    f"Positional args cannot be specified with "
                    f"options: {', '.join(options)}"
                )

        if action is None:
            if container:
                if is_bool_or:
                    action = BoolOrContainerAction.make(container, type)
                    # XXX: Type conversion handled in action
                    type = str
                else:
                    action = ContainerAction.make(container)
            elif is_bool:
                action = "store_true"
            elif is_bool_or:
                action = BoolOrAction

        if nargs is None:
            if is_positional:
                if container:
                    nargs = "+"
                elif is_optional:
                    nargs = "?"
            elif is_var_positional:
                nargs = "*"
            elif is_bool_or:
                if container:
                    nargs = "*"
                else:
                    nargs = "?"
            elif is_optional:
                if container:
                    nargs = "*"

        options = tuple(opt for opt in (short_option, long_option) if opt is not None)
        all_options = options

        if no_inverse:
            inverse_options = ()
        else:
            inverse_options = tuple(
                opt
                for opt in (inverse_short_option, inverse_long_option)
                if opt is not None
            )
            all_options += inverse_options

        if is_var_positional and default is EMPTY:
            default = ()

        self.command = command
        self.parameter = parameter
        self.is_positional = is_positional
        self.is_var_positional = is_var_positional
        self.is_optional = is_optional
        self.takes_value = is_positional or (is_optional and not is_bool)
        self.dest = parameter.name
        self.name = name
        self.metavar = metavar
        self.container = container
        self.type = type
        self.is_bool = is_bool
        self.is_bool_or = is_bool_or
        self.default = default
        self.choices = choices
        self.help = help
        self.inverse_help = inverse_help
        self.short_option = short_option
        self.long_option = long_option
        self.options = options
        self.no_inverse = no_inverse
        self.inverse_short_option = inverse_short_option
        self.inverse_long_option = inverse_long_option
        self.inverse_options = inverse_options
        self.all_options = all_options
        self.action = action
        self.nargs = nargs
        self.mutual_exclusion_group = mutual_exclusion_group
        self.envvar = envvar

    @cached_property
    def add_argument_args(self, *, _type_wrapper_cache={}):
        args = self.options
        if self.is_optional and not self.is_bool:
            if self.type not in _type_wrapper_cache:
                type = lambda v: None if v == "" else self.type(v)  # noqa: E731
                type = update_wrapper(type, self.type)
                _type_wrapper_cache[self.type] = type
            type = _type_wrapper_cache[self.type]
        else:
            type = self.type
        kwargs = {
            "action": self.action,
            "choices": self.choices,
            "dest": self.dest,
            "help": self.help,
            "metavar": self.metavar,
            "nargs": self.nargs,
            "type": type,
        }
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        if self.is_positional and self.is_optional:
            kwargs["default"] = PositionalPlaceholder
        return args, kwargs

    @cached_property
    def add_argument_inverse_args(self):
        if not (self.is_bool or self.is_bool_or) or self.no_inverse:
            return None

        args = self.inverse_options
        _, kwargs = self.add_argument_args

        if self.inverse_help:
            inverse_help = self.inverse_help
        elif self.help:
            inverse_help = invert_string(self.help)
        else:
            inverse_help = self.help

        kwargs = kwargs.copy()
        kwargs["action"] = "store_false"
        kwargs["help"] = inverse_help

        if self.is_bool_or:
            kwargs.pop("metavar")
            kwargs.pop("nargs")
            kwargs.pop("type")

        return args, kwargs

    def convert_value(self, value: str):
        """Convert string value to this arg's type."""
        if not isinstance(value, str):
            return value
        if self.is_bool or self.is_bool_or:
            if value in ("1", "true"):
                return True
            elif value in ("0", "false"):
                return False
            if self.is_bool:
                raise ValueError("Bool value must be one of 1, true, 0, or false")
        converter = self.add_argument_args[1]["type"]
        value = converter(value)
        return value

    def __str__(self):
        kind = "Positional" if self.is_positional else "Optional"
        has_default = self.default not in (EMPTY, None)
        default = f"[={self.default}]" if has_default else ""
        if self.is_bool:
            type = "flag"
        elif self.is_bool_or:
            type = f"flag|{self.type.__name__}"
        elif self.type is None:
            type = None
        else:
            type = self.type.__name__
        return f"{kind} arg: {self.name}{default}: type={type}"


class HelpArg(Arg):
    def __init__(self, *, command):
        parameter = Parameter(
            BaseParameter("help", POSITIONAL_OR_KEYWORD, default=False),
        )
        super().__init__(
            command=command,
            parameter=parameter,
            name="help",
            container=None,
            type=bool,
            positional=None,
            default=False,
            choices=None,
            help=None,
            inverse_help=None,
            short_option="-h",
            long_option="--help",
            no_inverse=True,
            inverse_short_option=None,
            inverse_long_option=None,
            action=None,
            nargs=None,
            mutual_exclusion_group=None,
            envvar=None,
        )


class bool_or:
    """Used to indicate that an arg can be a flag or an option.

    Use like this::

        @command
        def local(config, cmd, hide: {'type': bool_or} = False):
            "Run the specified command, possibly hiding its output.

            If ``hide=True``, *all* output will be hidden. It can also
            be set to one of "stdout" or "stderr" to hide just the
            specified output stream.

            "

    .. note:: The default inner type for ``bool_or`` is ``str``.
        ``bool_or(str)`` is equivalent to ``bool_or``.

    Allows for this::

        run local --hide all     # Hide everything
        run local --hide         # Hide everything with less effort
        run local --hide stdout  # Hide stdout only
        run local --no-hide      # Don't hide anything

    This can also be combined with the ``container`` option like so::

        @command
        def fetch(fields: {'container': dict, type: bool_or(int)}):
            "Get data from somewhere and show the specified fields.

            If ``fields=True``, all fields will be shown with their
            original names. If fields are specified, only those fields
            will be shown, with their names mapped. For example,
            ``fields`` could be::

                {'givenName': 'first_name', 'sn': 'last_name'}

            "

    .. note:: When combined with ``container``, the type passed to
        ``bool_or`` is applied to the *values* in the container rather
        than being applied to the literal strings passed on the command
        line.

    """

    type = str

    def __new__(cls, type, *, _type_cache={}):
        if type not in _type_cache:
            name = f"BoolOr{type.__name__.title()}"
            _type_cache[type] = builtins.type(name, (cls,), {"type": type})
        return _type_cache[type]


def add_items_to_container(
    container_type,
    item_type,
    existing_items,
    new_items,
    option_string,
):
    """Return a new container with existing plus new items."""
    items = []
    if is_mapping(existing_items):
        if existing_items is not None:
            items.extend(existing_items.items())
        for value in new_items:
            try:
                name, value = value.split(":", 1)
            except ValueError:
                raise CommandError(
                    f"Bad format for {option_string}; "
                    f"expected `name:<value>` but got `{value}`"
                )
            value = item_type(value)
            items.append((name, value))
        return container_type(items)
    elif is_sequence(existing_items):
        if existing_items is not None:
            items.extend(existing_items)
        items.extend(item_type(value) for value in new_items)
    else:
        raise ValueError(f"Not a mapping or sequence: {existing_items!r}")
    return container_type(items)


class BoolOrAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        if value is None:
            value = True
        setattr(namespace, self.dest, value)


class BoolOrContainerAction(argparse.Action):
    @classmethod
    def make(cls, container_type, item_type):
        return type(
            "BoolOrContainerAction",
            (cls,),
            {
                "container_type": container_type,
                "item_type": item_type,
            },
        )

    def __call__(self, parser, namespace, value, option_string=None):
        if value == []:
            setattr(namespace, self.dest, True)
        else:
            existing_items = getattr(namespace, self.dest)
            if isinstance(existing_items, bool):
                # XXX: The default value is False
                existing_items = self.container_type()
            items = add_items_to_container(
                self.container_type,
                self.item_type,
                existing_items,
                value,
                option_string,
            )
            setattr(namespace, self.dest, items)


class ContainerAction(argparse.Action):
    @classmethod
    def make(cls, container_type):
        return type("ContainerAction", (cls,), {"container_type": container_type})

    def __call__(self, parser, namespace, values, option_string=None):
        items = add_items_to_container(
            self.container_type,
            self.type,
            getattr(namespace, self.dest, self.container_type()),
            values,
            option_string,
        )
        setattr(namespace, self.dest, items)


def json_value(string):
    """Convert string to JSON if possible; otherwise, return as is."""
    try:
        string = json.loads(string)
    except ValueError:
        pass
    return string
